/*
 *
 * logat -- predlani vsehchny floatu ne zdrojaku na logaty
 *    
 */


/*
 * dokumentace:
 */


/* struct t_boxlik
 * ---------------
 *
 * pro ukladani naparsovaneho c'cka slouzi typ `t_boxlik' 
 * struct
 * 
 *
 * typedef struct t_boxlik
 * {
 *   int typ;  -- o jakou parsovaci jednotku jde  viz enum e_BX
 *                pokud je nastaven bit BX_LOGAT_BIT jde o binarni
 *                operator, ktery nahrazen "logatni" fci
 *
 *   int cast; -- zdali bude tento bude obklopen konverzi do/z logatu
 *                mozne hodnoty BX_NIC, BX_DTOL (double->logat),
 *                BX_ITOL(int->logat), BX_LTOI(logat->int)
 *                viz emum e_bxcast
 *
 *    struct t_boxlik *exprtyp; -- pro vyrazy: datovy typ vyrazu
 *
 *    struct t_boxlik *fst, *snd, *trd;
 *                prvni, truhy, treti potomek v syntaktickem stromu 
 *
 * pro boxliky odpovidajici usekum kusu zdrojaku
 *
 *    char *txt; -- kus zdrojaku
 *
 *    struct t_boxlik *next; -- nasledujici zdrojaku kus zdrojaku
 * } t_boxlik;
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <stdarg.h>
#include "yy.h"
#include "logat.h"

/*
 * pole nazvu konstant enum e_BX; velmi vhodne pro ladeni
 */
char* bx_nazvy[]=
{
 "UNKNOWN",

 "bx_un_exlaim",
 "bx_un_tilda",
 "bx_inc_pref",
 "bx_inc_post",
 "bx_dec_pref",
 "bx_dec_post",
 "bx_un_plus",
 "bx_un_minus",
 "bx_cast",
 "bx_asterix",
 "bx_obelix",
 "bx_sizeof_expr",
 "bx_sizeof_type",

 "bx_fce",
 "bx_array",
 "bx_ptr_op",
 "bx_dot",

 "bx_mul",
 "bx_div",
 "bx_mod",
 "bx_add",
 "bx_sub",
 "bx_lshift",
 "bx_rshift",
 "bx_lt",
 "bx_le",
 "bx_gt",
 "bx_ge",
 "bx_eq",
 "bx_neq",
 "bx_and",
 "bx_xor",
 "bx_or",
 "bx_land",
 "bx_lor",

 "bx_assig",
 "bx_assig_eq",
 "bx_assig_add",
 "bx_assig_sub",
 "bx_assig_mul",
 "bx_assig_div",
 "bx_assig_mod",
 "bx_assig_left",
 "bx_assig_right",
 "bx_assig_and",
 "bx_assig_xor",
 "bx_assig_or",
 "bx_comma",
 "bx_cond",

 "bx_txt",
 "bx_unary",
 "bx_fce_def",
 "bx_def_list",
 "bx_expr_list",
 "bx_arg_list",
 "bx_max",
 "bx_declaration",
 "bx_declaration_spec",
 "bx_init_declarator",

 "bx_expr",
 "bx_tok_typedef",
 "bx_tok_extern",
 "bx_tok_static",
 "bx_tok_auto",
 "bx_tok_register",
 "bx_tok_char",
 "bx_tok_short",
 "bx_tok_int",
 "bx_tok_long",
 "bx_tok_signed",
 "bx_tok_unsigned",
 "bx_tok_float",
 "bx_tok_double",
 "bx_tok_const",
 "bx_tok_volatile",
 "bx_param_list",
 "bx_struct_decl_list",
 "bx_struct_or_union_spec",
 "bx_tok_struct",
 "bx_tok_union",
 "bx_struct_decl",
 "bx_struct_declarator",
 "bx_enum_spec",
 "bx_enumerator",
 "bx_declarator",
 "bx_decl2_inpar",
 "bx_decl2_bra",
 "bx_decl2_fce",
 "bx_pointer",
 "bx_elipsis",
 "bx_id_decl",
 "bx_par_decl",
 "bx_adeclarator",
 "bx_initializer",
 "bx_statement",
 "bx_compound",
 "bx_if",
 "bx_fce_body",
 "bx_declspec",
 "bx_type_pointer",
 "bx_type_fce",
 "bx_identifier",
 "bx_end",
 "bx_type_bit",
 "void",
 "char",
 "short",
 "int",
 "long",
 "llong",
 "signed",
 "unsigned",
 "float",
 "double",
 "logat", 
 "const",
 "volatile",
 "bx_type_name",
};

/*
 * nazvy fci pro konverzi logat<->float,double
 */
char *bx_nazvy_cast[]=
{
  "nic(*&%",  /* nemelo by se nikdy obevit */

  "lg_itol(", /* int --> logat */
  "lg_ltoi(", /* logat --> int */

  "lg_ltod(", /* logat  --> doble */
  "lg_dtol(", /* double --> logat */

}; 


/*
 * koren syntaktickeho stromu
 */
t_boxlik *boxlik_root= NULL;

/*
 * uplne prvni boxlik boxliku zdrojaku, (pospojovane pomoci `next')
  */
t_boxlik *fst_boxlik=NULL;

/*
 * privni boxlik tehoz spojaku 
 */
t_boxlik *lst_boxlik=NULL;

/*
 * indikator jestli se parser nachazi ve fci 
 * a deklarace jsou pouze lokalni ci globalni
 */
int bx_infce=0;

/*
 * bool jestli se budou tisknout tabulky promenych
 * (lokalni pred kazdym vyprazdnenim
 * globalni na konci
 */

int print_var_table = 0 ;


/*
 * vystupni a errorove file, co dotat?
 */
FILE *bx_file_out = NULL;
FILE *bx_file_err = NULL;

void idtable_print(t_idtable *t);
void print_tree_one_data_typ(t_boxlik *bx);

/*******************
 * id parser helpers
 */

void
bx_printf(char *fmt,...)
{
  va_list ap;

  if( bx_file_out )
  {
    va_start(ap, fmt);
    vfprintf(bx_file_out, fmt, ap);
    va_end(ap);
  }
  
} /* bx_printf */

void
bx_err(char *fmt,...)
{
  va_list ap;

  va_start(ap, fmt);
  vfprintf(bx_file_err, fmt, ap);
  va_end(ap);
  
}
/*
 * asi na nic
 */
int bx_isfloat(t_boxlik *bx)
{
  if( bx == NULL )
  {
    return 0;
  }
  if (bx->typ == BX_TOK_FLOAT )
  {
    return 1;
  }
  return 0;
} /* bx_isfloat */

/*
 * pridava do id typ ktery je spolecny pro cely list identifikatoru
 *
 *   int * a, aa[8],  aa; // v tomto pripade ``int''
 *   id->fst doposud naparsovane 
 *   add typ co se ma pridat
 *   `id->snd' ukazuje na posledni polozku kam je treba tento typ pridat
 */
void
bx_id_truetype(t_boxlik *id, t_boxlik *add)
{
  assert(id);
  //assert(id->typ == BX_IDENTIFIER);
  assert(id->typ == BX_IDENTIFIER || id->typ == BX_ADECLARATOR);



  if( id->snd == NULL )
  {
    id->fst = add;
    return;
  }

  assert(id->snd->typ == BX_POINTER  || id->snd->typ == BX_TYPE_FCE);

  id->snd->fst = add;

  /* id->snd = NULL; */

} /* bx_id_truetype */

/*
 * predhodi poitery
 */
void
bx_id_addpoiter(t_boxlik *id, t_boxlik *ptr)
{
  t_boxlik *last;

  assert(id);
  assert(id->typ == BX_IDENTIFIER || id->typ == BX_ADECLARATOR);

  assert(ptr);
  assert(ptr->typ == BX_POINTER);

    
  /* dojdi az na konec seznamu */
  for(last = ptr; last->fst; last = last->fst);


  if( id->fst == NULL || id->fst->typ != BX_TYPE_FCE )
  { /* predpoji se */
    last->fst = id ->fst;

    id->fst = ptr;

    if( id ->snd == NULL ) 
    {
      id->snd = last;
    }
  }
  else
  { /* je to fce */
    assert(id->fst->fst == NULL);
    id->fst->fst = ptr;
    id->snd = last;
  }
    
} /* bx_id_addpoiter */


/*
 * predhodi typ identifikatoru. [id] -> typ
 *                                 `>typ-^
 */
void bx_id_addtype(t_boxlik *id, t_boxlik *add)
{
  assert(id);
  assert(id->typ == BX_IDENTIFIER);


  if( id->fst == NULL ) 
  {
    id->fst = add;
  }

  if( id ->snd == NULL ) 
  {
    id->snd = add;
  }

} /* bx_id_addtype */

/*
 * prida promenou do lokalni tabulky
 * predpoklada ze promena je jiz otypovana (id->fst je typ)
 */
void
bx_id_addlocally(t_boxlik *id)
{
  t_boxlik *type;

  assert(id);
  assert(id->typ == BX_IDENTIFIER);
  assert(id->fst);
  assert(id->txt);

  type = id->fst;

  /* FIXME: very very *very* dirty */
  /* kdyz je prvni const tak se preskoci */

#if 1
  if( type->typ == BX_TOK_CONST )
  {
    type = type->snd;
    assert(type);
  }
#endif

  assert(type->typ > BX_TYPE_BIT ||
         type->typ == BX_STRUCT_OR_UNION_SPEC ||
         type->typ == BX_POINTER ||
         type->typ == BX_ENUM_SPEC);
  

  idtable_add(type_tbl_loc, id->txt, type, NULL);
} /* bx_id_addlocally */


/*
 * syntakticky strom vytvori deklaratoru typ, akorat mu chybi
 * specifier. ten muze byt jede pro cely list deklaratoru.
 * tato fce vytvori ucelene typy
 */
void bx_id_addspecifier(t_boxlik *specifier, t_boxlik *declarator)
{
  t_boxlik *identifier;
  t_boxlik *type, *storage;

  t_idtable *table;

  assert(declarator);
  assert(specifier);

  storage = specifier->fst;

  type = specifier->snd;
  assert(type);
  assert(type->typ > BX_TYPE_BIT ||
         type->typ == BX_STRUCT_OR_UNION_SPEC ||
         type->typ == BX_ENUM_SPEC);

  table = bx_infce ? type_tbl_loc :  type_tbl;

  for(;declarator; declarator = declarator->snd)
  {
    identifier = declarator->fst;
    assert(identifier);
    assert(identifier->typ == BX_IDENTIFIER);
   
    bx_id_truetype(identifier, type);
    identifier->next = NULL;
    idtable_add(table, identifier->txt, identifier->fst, storage);
  }
} /* bx_id_addspec */


    
/**************
 * parser helpers
 */

/*
 * pro syntakticka pravidla typu
 * expr_list
 *   : expr
 *   | expr_list ',' expr
 *
 *
 *   se (v druhem pripade) az nakones seznamu (tvoreny pomoci `snd') 
 *   na `fst' se pak vesi jednotlive `expr'
 */
void bx_last(t_boxlik* ss, t_boxlik* last)
{
  assert(ss != NULL);

  /* dojed na konec */
  for(;ss->snd;ss= ss->snd);

  /* pripoj */
  ss->snd = last;

} /* bx_last */




/*
 * nejak inteligentne udela BX_DECLSPEC 
 * dopise do *decl* storage a type, pokud neexistujne vytvori ho.
 */
t_boxlik* bx_declspec(t_boxlik* storage, t_boxlik* type, t_boxlik* decl)
{
  if( decl == NULL )
  {
    decl = boxlik_new(BX_DECLARATION_SPEC, NULL, NULL, NULL);
    decl->snd = BX0(BX_TYPE_BIT);
  }

  if( storage != NULL )
  { /* pridavame storage */
    if( decl->fst == NULL )
    { /* misto na storage je prazdne */
      decl->fst = storage;
    }
    else
    { /* neni misto pridame do seznamu storagu */
      bx_last(decl->fst, storage);
    }
  }

  if( type != NULL )
  { /* pridavame typ */
    if(type->typ == BX_STRUCT_OR_UNION_SPEC ||
       type->typ == BX_ENUM_SPEC ||
       type->typ == BX_TYPE_NAME)
    {
      /* delete decl->snd */
      decl->snd = type;
    }
    else
    {
      assert(type->typ > BX_TYPE_BIT);
      decl->snd->typ += (type->typ & (~BX_TYPE_BIT) );
    }
  }

  return decl;

} /* bx_declspec */


/*
 * prohleda BX_DECLARATION_SPEC jestli timto definujeme novy typ
 * tj zdali obsahuje typedef
 */
int bx_istypedefstorage(t_boxlik* decl)
{
  t_boxlik* storage;

  if(decl == NULL)
  {
    return 0;
  }

  storage = decl->fst;

  if( storage == NULL )
  {
    return 0;
  }

  /* pouze jedna storage ne? */

  if( storage->typ == BX_TOK_TYPEDEF )
  {
    return !0;
  }
  return 0;
} /* bx_istypedefstorage */

/*
 *
 * asi nepouzito
 *
 * prida typ do tabulky typu
 * FIXME: only one type in list
 */

void bx_addtype(t_boxlik* declaration)
{

  t_boxlik *declspec, *ideclarator, *declarator, *identifier;

return;

  assert(declaration != NULL);
  assert(declaration->typ == BX_DECLARATION);

  declspec = declaration->fst;
  assert(declspec);
  assert(declspec->typ == BX_DECLARATION_SPEC);

  if( !bx_istypedefstorage(declspec) )
  { /* nejde o deklaraci typu */
     return ;
  }

 
  ideclarator = declaration->snd;
  assert(ideclarator);
  assert(ideclarator->typ == BX_INIT_DECLARATOR);

  declarator = ideclarator->fst;
  assert(declarator);
  assert(declarator->typ == BX_DECLARATOR);

  identifier = declarator->snd;
  assert(identifier);
  assert(identifier->typ == BX_TXT);

  idtable_add(type_tbl, identifier->txt, NULL, NULL);

} /* bx_addtype,asi na nic */

/**************
 * id tables, nafukovaci tabulka identifikatoru
 *
 * zadas `id' (char*), a vypadne ti typ(t_boxlik*)
 * 
 * to vse pomoci toho interfacu:
 *
 * t_idtable *idtable_new() vytvori novou tabulku po
 * 
 * t_id_item *idtable_find(t, char*key) najdi v `t' identifikator `key'
 *
 * idtable_hastype(t_idtable *t, char *key) je v `t' identifikator `key'?
 *
 * idtable_add(t, key, typ, storage) pridej do `t' pod nazvem `key'
 *    typ `typ'(t_boxlik*) storage (const|typedef|...) 
 * 
 * idtable_clean(t);
 * 
 * a to staci ne?
 * 
 * 
 */
#define IDTBL_SIZ 30
void print_tree_one_type(t_boxlik *bx);

t_idtable *type_tbl = NULL;
t_idtable *type_tbl_loc = NULL;

t_idtable *
idtable_new()
{
  t_idtable *ret;

  ret = (t_idtable*) malloc(sizeof(t_idtable));
  ret -> top = 0;
  ret -> siz= IDTBL_SIZ;
  ret -> tbl = (t_id_item**) malloc(ret -> siz * sizeof(t_id_item*));

  assert(ret);

  return ret;
} /* idtable_new */

void
idtable_clean(t_idtable *t)
{
  if( print_var_table )
  {
    bx_printf("---== type_tbl_loc ==---\n\n");
    idtable_print(type_tbl_loc);
  }
  /* FIXME: very dirty */
  t->top = 0;
} /* idtable_clean */

int bx_hastypedef(t_boxlik *storage)
{
  for(;storage; storage = storage->next)
  {
    if( storage -> typ == BX_TOK_TYPEDEF )
    {
      return 1;
    }
  }
  return 0;
} /* bx_hastypedef */

t_id_item* 
idtable_find(t_idtable *t, char *key)
{
  int i;

  for(i = t->top; i--; )
  {
    if( strcmp (t->tbl[i]->id, key) == 0 )
    {
      return t->tbl[i];
    }
  }
  return NULL;

} /* idtable_has */

/*
 * nastavi typ boxliku *id* dle toho jaky typ najde v tabulkach 
 */
t_boxlik *bx_idtype(t_boxlik *id)
{

  t_id_item *item;
  assert(id);
  assert(id->typ == BX_TXT);

  item = idtable_find(type_tbl_loc, id->txt);

  if( item == NULL )
  { /* jeste neni: zadne opicky */
    item = idtable_find(type_tbl, id->txt);
    if( item == NULL )
    {
      return;
    }
  }

  id->exprtyp = item->typ;
  
} /* idtable_idtype */

int
idtable_hastype(t_idtable *t, char *key)
{
  t_id_item *item;

  assert(t);

  item = idtable_find(t, key);

  if( item == NULL )
  {
    return 0;
  }

  if( bx_hastypedef(item->storage) )
  {
    return 1;
  }
  return 0;

} /* idtable_hastype */


int
idtable_has(t_idtable *t, char *key)
{
  int i;

  assert(t);

  for(i = t->top; i--; )
  {
    if( strcmp (t->tbl[i]->id, key) == 0 )
    {
      return 1;
    }
  }

  return 0;
} /* idtable_has */

void
idtable_add(t_idtable *t, char *key, t_boxlik *typ, t_boxlik *storage)
{
  assert(t);
  if( idtable_has(t, key) )
  {
    bx_err("juz defined(%s)", key);
    /* FIXME: tiskne na bx_file_out, misto na bx_file_err */
    print_tree_one_type(typ);
    bx_err("\n");
    return;
  }
  assert(t->siz > 0);
  if(  t->top > t->siz )
  { /*zvetsit tabulku */
    t->siz *= 2;
    t->tbl = (t_id_item**)realloc(t->tbl, t->siz * sizeof(t_id_item*));
  }

  t->tbl[t->top]= id_item_new(key, typ, storage);
  t->top ++;

} /* idtable_add */

t_id_item*
id_item_new(char *key, t_boxlik *typ, t_boxlik *storage)
{
  t_id_item *ret;

  ret = (t_id_item*)malloc(sizeof(t_id_item));
  ret->id = strdup(key);
  ret->typ = typ;
  ret->storage = storage;

  return ret;

} /* id_item_new */

/*
 * tisk tabulky, pro ladici ucely
 */

void
idtable_print(t_idtable *t)
{
  int i;
  assert(t);
  for(i = 0; i < t->top; i++)
  {
    bx_printf("%2i %3s[%i] ",i , t->tbl[i]->id, bx_hastypedef(t->tbl[i]->storage));
    print_tree_one_type(t->tbl[i]->typ);
    bx_printf("\n");
  }
} /* idtable_print */


/**************
 * boxlik
 *
 * nejake fce na ovladani boxliku
 * 
 */
t_boxlik *boxlik_new(int typ, t_boxlik *fst, t_boxlik *snd, t_boxlik *trd)
{
  t_boxlik *ret;

  ret= (t_boxlik*) malloc(sizeof(t_boxlik));


  ret->fst = fst;
  ret->snd = snd;
  ret->trd = trd;
  ret->typ = typ;
  ret->exprtyp = NULL;
  ret->txt = NULL;
  ret->next = NULL;
  ret->cast = BX_NIC;
  return ret;

} /* boxlik_new */


/*
 * inter face k boxlik_new: vytvori prazdny boxlik obsahujici akorat `txt'
 */
t_boxlik*
boxlik_new_txt(char *txt)
{
  t_boxlik *ret;

  ret = boxlik_new(BX_TXT, NULL, NULL, NULL);

  ret->txt = strdup(txt);

  return ret;

} /* t_boxlik_new */

/*
 * vytvareni rady boxliku zdrojaku se deje nasledujici fci
 * z `txt' se udela boxlik a prida se na konec seznamu
 * zacinajiciho v `fst_boxlik' a konciciho v `lst_boxlik'
 *
 */
t_boxlik*
boxlik_add(char *txt)
{
  t_boxlik *ret;

  ret = boxlik_new_txt(txt);

  if(lst_boxlik)
  { /* pripoj na konec */
    lst_boxlik->next= ret;
    lst_boxlik= ret;
  }
  else
  { /* je to prvni boxlik */
    lst_boxlik =
    fst_boxlik = ret;
  }

  return ret;

} /* boxlik_add */

/*********************************************
 *
 * bx type expresions proces
 *
 * ohodnocovani typu 
 */
int bxt_isptr(t_boxlik *bx)
{
  assert(bx);
  assert(bx->exprtyp);

  return bx->exprtyp->typ == BX_POINTER;
} /* bxt_isptr */

/*
 * neexistuje takove makro v standartnich headre ?
 * vitas nevi
 */
#define BXMAX(a, b) ((a)<(b)?(b):(a))


/*
 * pokud typ a je typedefem nejakeho jineho preloz
 */
t_boxlik *
bxt_odtypedef(t_boxlik *a)
{
  
  t_id_item *it;
  assert(a);
  if( a-> typ == BX_TYPE_NAME )
  { /* najdi v globalni tabulce definic */
    assert(a->fst);
    assert(a->fst->txt);
    it = idtable_find(type_tbl, a->fst->txt);
    assert(it);
    return it->typ; 
  }
  return a;
} /* bxt_odtypedef */

/*
 * pokud do expesionu vstupuji dva typy, tohle by melo 
 * urcit jejich vysledek
 * 
 * vychazi se s posloupnosti (vetsi bere)
 * char < int < long < float < logat ==  double
 *
 * pokud je jeden z typu pointer bere se typ toho ktery je drive
 * 
 */
t_boxlik *bxt_resulttyp(t_boxlik *a, t_boxlik *b)
{
  int res;

  t_boxlik *aa,  *bb;

  aa = a->exprtyp;
  bb = b->exprtyp;

  assert(aa);
  assert(bb);

  aa = bxt_odtypedef(aa);
  bb = bxt_odtypedef(bb);

  if( bxt_isptr(a))
  {
    return aa;
  }

  if( bxt_isptr(b))
  {
    return bb;
  }

  assert(aa->typ > BX_TYPE_BIT);  
  assert(bb->typ > BX_TYPE_BIT);  
  assert(aa->typ <= BX_TOK_LOGAT);
  assert(bb->typ <= BX_TOK_LOGAT);

  if( aa->typ  > BX_TOK_FLOAT || bb->typ > BX_TOK_FLOAT )
  {
    res =  BX_TOK_LOGAT;
  }
  else
  {
     res = BXMAX(aa->typ, bb->typ);
  }

  return BX0(res);
} /* bxt_restyp */


/*
 * nacastuje `a' podle toho jakou ma hodnotu `ss'
 */

void bxt_cast_by_example(t_boxlik *boxcto, t_boxlik *a)
{
  int cfrom, cto;

  assert(a->exprtyp);

  /* cto =  ss->exprtyp->typ; */
  cto = boxcto->typ;
  cfrom = a->exprtyp->typ;

  if( cto == BX_TOK_LOGAT )
  { /* cil je logat */
	  
    if( cfrom == BX_TOK_LOGAT )
    { /* oba sou to logat, tak nezmatkuj */
      return;
    }

    if( cfrom >= BX_TOK_FLOAT )
    { /* from je float a nebo vice */
      a->cast = BX_DTOL;
      return;
    }

    assert(cfrom >= BX_TYPE_BIT);
    /* from je int  (tj. void az long) */

    a->cast = BX_ITOL;

    return;
  }

  /* cto neni logat */

  if( cfrom != BX_TOK_LOGAT )
  { /* from taky neni logat, tak se stim nebudem parat
     * compilator si stim poradi
     */
    return;
  }
  
  /* cfrom je logat */
  if( cto >= BX_TOK_FLOAT )
  {
    a->cast = BX_LTOD;
    return;
  }
   
  assert(cfrom >= BX_TYPE_BIT);

  /* cto je int  (tj. void az long) */
  a->cast = BX_LTOI;

} /* bxt_cast_by_example */

/*
 * v pripade, ze `ss' je logat bude pouzit logatova fce (tj nastavi se
 * BX_LOGAT_BIN
 */

int
bxt_logatop(t_boxlik *ss)
{
  assert(ss);
  assert(ss->exprtyp);

  assert(ss->typ >= BX_UN_EXLAIM && ss->typ <= BX_COND );


  if( ss->exprtyp->typ == BX_TOK_LOGAT )
  {
    ss->typ |= BX_LOGAT_BIT;
  }
} /* bxt_logatop */

/*
 * klasicky binop +,*,-,/,%,...
 */

t_boxlik *bxt_binop(t_boxlik *ss, t_boxlik *a, t_boxlik *b)
{
  t_boxlik *restyp;

  restyp = bxt_resulttyp(a, b);

  ss->exprtyp = restyp;

  bxt_cast_by_example(ss->exprtyp, a);
  bxt_cast_by_example(ss->exprtyp, b);


  bxt_logatop(ss);
} /* bxt_binop */


/*
 * kopirovani typu z `a' do `ss'
 */
void bxt_cptyp  (t_boxlik *ss, t_boxlik *a)
{
  assert(a);
  assert(ss);
//  assert(a->exprtyp);

  ss->exprtyp = a->exprtyp;

} /* bxt_cptyp */ 

/*
 * pokud je identifikator deklarovan, ma next jiny vyznam.
 * totiz: ukazuje kde byl naposled napsan typ.
 */
t_boxlik*
bxt_identifier_decl(t_boxlik *id)
{
  t_boxlik *ret; 
  assert(id->typ == BX_TXT);
  assert(id->txt);

  ret = BX0(BX_IDENTIFIER);
  ret->txt = strdup(id->txt);
  ret->next = ret;

  return ret;
} /* bxt_identifier_decl */

/*
 * timto timto zpusobem je deklarovana nova promena
 * rovnou se ji priplacne typ.
 */ 
void
bxt_identifier(t_boxlik *ss, t_boxlik *a)
{
  t_id_item *it;
  t_boxlik *typ;

  assert(a);
  assert(a->txt);
  assert(a->typ == BX_TXT);

  /* hledej v lokalni a globalni tabulce promenych */
  it = idtable_find(type_tbl_loc, a->txt);
  if( it == NULL )
  {
    it = idtable_find(type_tbl, a->txt);
  }

  if(!it)
  {
    bx_err("cannt find %s\n", a->txt);
    
  }
  assert(it);

  typ = it->typ;
  assert(typ);

  if( typ->typ == BX_TYPE_NAME )
  { /* pokud je to jen nazev typu expanduj, tady by mohlo byt while  */
    assert(typ->fst);
    assert(typ->fst->txt);
    it = idtable_find(type_tbl_loc, typ->fst->txt);
    if( it == NULL )
    {
      it = idtable_find(type_tbl, typ->fst->txt);
    }
    assert(it);
  }

  ss->exprtyp = it->typ;

  ss->txt = a->txt;
  ss->fst = a;
   
} /* bxt_identifier */

/*
 * pretipovani na jiny typ 
 */

void
bxt_cast( t_boxlik *ss, t_boxlik *a, t_boxlik *b)
{
  assert(ss);
  assert(a);
  assert(b);

  assert(a->typ == BX_ADECLARATOR);

  assert(a->fst);

  ss->exprtyp = a->fst;

} /* bxt_cast */

/*
 * typ prirazeni je vzdy podle lval, je treba pretypovat veci na prave strane
 */
void bxt_assig  (t_boxlik *ss)
{

  t_boxlik *a = ss->fst;
  t_boxlik *b = ss->snd;
  t_boxlik *c = ss->trd;
  /*
   * lval = expr
   *  a   b  c
   */
  assert(ss);
  assert(a);
  assert(c);

  bxt_cptyp(ss, a);
  bxt_cast_by_example(ss->exprtyp, c);

} /* bxt_assig */

/*
 * obcas se hodi neco pretypovat na int
 */
void bxt_toint  (t_boxlik *a)
{
  assert(a);
  assert(a->exprtyp);

  if( a->exprtyp->typ == BX_TOK_LOGAT )
  {
    a->cast = BX_ITOL;
  }

  a->exprtyp = BX0(BX_TOK_INT);

} /* bxt_toint */


void bxt_constant(t_boxlik *ss)
{
  assert(ss);

  ss->exprtyp = BX0(BX_TOK_FLOAT);
} /* bxt_constant */

/*
 * unary operatory
 */

void
bxt_unary( t_boxlik *ss, t_boxlik *a)
{
  t_boxlik *op;

  assert(ss);
  assert(a);

  assert(a -> exprtyp);

  assert(ss->fst);
  assert(ss->fst->typ == BX_TXT);

  op = ss->fst;

  if( 0 == strcmp("*", op->txt) )
  { /* resime co s dereferenci */
    assert(a -> exprtyp -> typ == BX_POINTER);
    ss-> exprtyp = a ->exprtyp -> fst;
    return;
  }

  if( 0 == strcmp("&", op->txt) )
  { /* resime co s dereferenci */
    ss-> exprtyp = BX1(BX_POINTER, a ->exprtyp -> fst);
    return;
  }
} /* bxt_unary */

/*
 * bxt_array []
 */
void
bxt_array(t_boxlik *ss, t_boxlik *a, t_boxlik *b)
{
  assert(ss);
  assert(a);
  assert(b);

  assert(a->exprtyp);
  assert(a->exprtyp->typ == BX_POINTER);

  ss-> exprtyp = bxt_odtypedef( a ->exprtyp -> fst );
  bxt_toint(b);

} /* bxt_array */

/*
 * prefixove ++, --
 */
void
bxt_preincdec(t_boxlik *ss, t_boxlik *a)
{
  assert(ss);
  assert(a);


  bxt_cptyp(ss, a);
  /* FIXME: logat ++ */
  /* bxt_logatop(a); */
} /* bxt_preincdec */

void bxt_land   (t_boxlik *ss, t_boxlik *a, t_boxlik *b)
{
} /* bxt_land */

/*
 *  expr?expr:expr
 */
void bxt_cond   (t_boxlik *ss)
{
} /* bxt_cond */


/*
 * >>,<<
 */
void bxt_shift  (t_boxlik *ss, t_boxlik *a, t_boxlik *b)
{
} /* bxt_shift */

/*
 * dereference *
 */
void bxt_asterix(t_boxlik *ss, t_boxlik *a)
{
} /* bxt_asterix */

/*
 * volani fce
 */
void bxt_fce    (t_boxlik *ss, t_boxlik *a, t_boxlik *b)
{
  t_id_item *it;
  t_boxlik *rettyp;
  t_boxlik *params;


  /* nejprve vyresime return typ */

  assert(a->typ == BX_IDENTIFIER );
  assert(a->fst->typ == BX_TXT);

  it = idtable_find(type_tbl, a->fst->txt);

  assert(it->typ);
  assert(it->typ->typ == BX_TYPE_FCE );

  rettyp = it->typ->fst;
  params = it->typ->snd;

  if( rettyp->typ == BX_TYPE_NAME )
  { /* pokud je to jen nazev typu expanduj, tady by mohlo byt while  */
    assert(rettyp->fst);
    assert(rettyp->fst->txt);
    it = idtable_find(type_tbl, rettyp->fst->txt);
    assert(it);
    rettyp = it->typ;
  }

  ss->exprtyp = rettyp;

  ss->txt = a->fst->txt;
  ss->fst = a;

  /* mno a ted argumety */

  /*  */

  for(; b; b = b->fst, params = params -> fst)
  {
    /* ted jen prekonvertovat b do params */
    assert(b->typ ==  BX_ARG_LIST);
    assert(b->snd);

    assert(params);
    assert(params->typ == BX_PARAM_LIST);
    assert(params->snd);
    assert(params->snd->typ == BX_PAR_DECL || 
           params->snd->typ == BX_ADECLARATOR);
    assert(params->snd->fst);

    bxt_cast_by_example(params->snd->fst, b->snd);

  }

} /* bxt_fce */

/*
 * tecka bx.next
 */
void bxt_dot    (t_boxlik *ss, t_boxlik *a, t_boxlik *b)
{
  /* je treba nalezt typ vyrazu typu neco.necodal */
  t_boxlik *dl; /* struct_decl_list */
  t_boxlik *d; /* struct_decl */

  assert(b->typ == BX_TXT);
  assert(b->txt);
  
  assert(a);
  assert(a->exprtyp);
  assert(a->exprtyp->typ == BX_STRUCT_OR_UNION_SPEC );
  assert(a->exprtyp->trd);

  dl = a->exprtyp->trd;

  for(; dl; dl = dl->snd)
  {
    assert(dl->typ == BX_STRUCT_DECL_LIST);
    assert(dl->fst);
    assert(dl->fst->typ == BX_STRUCT_DECL);
    assert(dl->fst->snd);
    assert(dl->fst->snd->typ == BX_STRUCT_DECLARATOR);

    d = dl->fst->snd;
   
    for(;d; d = d->snd)
    {
      assert(d->fst);
      assert(d->fst->typ == BX_IDENTIFIER);
      assert(d->fst->txt);
      assert(d->fst->fst);

      bx_err("bxt_dot: nalezl jsem %s\n", d->fst->txt);
      if( 0 == strcmp(b->txt, d->fst->txt) )
      { /* vrat typ */
        ss->exprtyp = d->fst->fst;
	return;
      }
    }
  }

} /* bxt_dot */

/*
 * ptrop bx->next
 */
void bxt_ptrop  (t_boxlik *ss, t_boxlik *a, t_boxlik *b)
{
} /* bxt_ptrop */

/*
 * post fix ++, --
 */
void
bxt_incdec (t_boxlik *ss, t_boxlik *a)
{
  assert(ss);
  assert(a);

  bxt_cptyp(ss, a);
  /* bxt_logatop(a); */

} /* bxt_incdec */



/*****************
 *
 * tisk syntaktickeho stromu 
 * vylozene pro ladici ucely
 *
 */

char pretext[100];


/*
 * tisni jeden boxlik
 *
 * `-- [* LG_bx_nasev ]  { typ } `' 
 *   ^  ^ ^  ^           ^       ^-- jeli neco v `txt' vypise se do uvozovek
 *   |  | |  `-`typ'          `-- `exprtyp' poku neni null
 *   |  | `-- jeli nastaven BX_LOGAT_BIT
 *   |  `-- pokud je `next' nenulove
 *   `--z mohou byt znaky -,=,E dle toho zda jede o `fst', `snd', `trd'
 *
 * k tisku se vyuziva `pretext' ve ktere jsou sloupce 
 */
void
print_tree_one_type(t_boxlik *bx)
{

  int typ;
  int i;

  if(bx == NULL )
  {
    bx_printf("null");
    return;
  }


  typ = bx->typ;
  typ = typ < 0? 0 : typ;

  if( typ & BX_TYPE_BIT || typ == BX_TYPE_FCE || typ == BX_POINTER )
  {
    print_tree_one_data_typ(bx);
    return;
  }

  if( bx->next )
  {
    bx_printf("^");
  }

  if( typ & BX_LOGAT_BIT )
  {
     bx_printf("LG_");
     typ &= ~BX_LOGAT_BIT;
  }

  if(typ & BX_TYPE_BIT)
  {
    for(i=0; i < 16; i++, typ>>=1)
    {
      if( typ & 1)
      {
        bx_printf("%s ",bx_nazvy[BX_END + 2 + i]);
      }
    }

  }
  else
  {
    typ = typ > BX_END ? 0 : typ;
    bx_printf("%s", bx_nazvy[typ]);
  }


} /* print_tree_one_type */

void
print_tree_one_data_typ(t_boxlik *bx)
{
  int print_comma=0;
  int typ;
  int i;

  if(NULL == bx)
  {
     return;
  }
  
  if( bx->typ == BX_TYPE_NAME )
  {
    /* print_tree_one_data_typ(bx->fst); */
    assert(bx->fst);
    assert(bx->fst->typ == BX_TXT);
    assert(bx->fst->txt);
    bx_printf("%s", bx->fst->txt);
    return;
  }
    

  if( bx->typ == BX_POINTER )
  { /* pointer */
    bx_printf("*");
    print_tree_one_data_typ(bx->fst);
    return;
  }

  if( bx->typ == BX_TYPE_FCE )
  { /* funkce */
     print_tree_one_data_typ(bx->fst);
     bx_printf("(");
     print_comma = 0;
     for(bx = bx->snd;bx;bx = bx->fst)
     {
       if( print_comma )
       {
         bx_printf(",");
       }
       assert(bx->typ == BX_PARAM_LIST);
       assert(bx->snd);
       assert(bx->snd->typ == BX_PAR_DECL || bx->snd->typ == BX_ADECLARATOR);
       //assert(bx->fst->fst);
       //assert(bx->fst->fst->typ == BX_IDENTIFIER);
       print_tree_one_data_typ(bx->snd->fst);
       print_comma = 1;
     }
     bx_printf(")");
     return;
  }

  /* zakladni typ */
  typ = bx->typ;
  typ = typ < 0? 0 : typ;

  if( bx->next )
  {
    bx_printf("^");
  }

  if( typ & BX_LOGAT_BIT )
  {
     bx_printf("LG_");
     typ &= ~BX_LOGAT_BIT;
  }

  if(typ & BX_TYPE_BIT)
  {
    for(i=0; i < 16; i++, typ>>=1)
    {
      if( typ & 1)
      {
        bx_printf(" %s",bx_nazvy[BX_END + 2 + i]);
      }
    }

  }
  else
  {
    typ = typ > BX_END ? 0 : typ;
    bx_printf("%s", bx_nazvy[typ]);
  }


} /* print_tree_one_data_typ */

void 
print_boxlik_tree_one(int dep,  t_boxlik *box, int isntlast, char face)
{
  int typ;
  int i;
  

  pretext[dep] = '\0';

  bx_printf("%s", pretext);
  if( isntlast)
  {
     bx_printf("+-");
     pretext[dep] = '|';
  }
  else
  {
    bx_printf("`-");
    pretext[dep] = ' ';
  }
  bx_printf("%c", face);
   

  pretext[dep+1] = '\0';

  
  bx_printf("[ ");
  print_tree_one_type(box);
  bx_printf("] ");

  if( box->exprtyp != NULL )
  {
    bx_printf("{ ");
    print_tree_one_type(box->exprtyp);
    bx_printf("} ");
  }

  if( box->cast != 0 )
  {
    bx_printf("(%s)", bx_nazvy_cast[box->cast]);
  }

  if( box->txt )
  {
    bx_printf(" `%s'", box->txt);
  }
  bx_printf("\n");

} /* print_boxlik_tree_one */

/*
 * rekurzivni tisk syntaktickeho stromu
 */
void
print_boxlik_tree(int dep, t_boxlik *box, int islast, char face)
{
 
  print_boxlik_tree_one(dep, box, islast, face);

#if 0
  if(box->typ == BX_TXT)
  {
    return;
  }
#endif

  dep++;

  if( box->fst)
  {
    print_boxlik_tree(dep, box->fst, box->snd||box->trd, '-');
  }
  if( box->snd)
  {
    print_boxlik_tree(dep, box->snd, !!box->trd, '=');
  }
  if( box->trd)
  {
    print_boxlik_tree(dep, box->trd, 0, 'E');
  }

} /* print_boxlik_tree */


/*************
 *
 * tist pomoci pruchodu stromem, uz se nepouziva
 * jsou pouzity pouze `leftist()' a `rightist()'
 */

void
print_boxlik_pp_one(t_boxlik *box)
{
  if( box->txt )
  {
    bx_printf("%s", box->txt);
  }
} /* print_boxlik_pp_one */
#if 0

t_boxlik *leftist(t_boxlik *box)
{
  if( box == NULL )
  {
    return NULL;
  }

  ret = leftist(box->fst);
  if( ret != NULL)
  {
    return ret;
  }

  ret = leftist(box->snd);
  if( ret != NULL)
  {
    return ret;
  }
  
  ret = leftist(box->trd);
  if( ret != NULL)
  {
    return ret;
  }

  if( box->typ == BX_TXT )
  {
    return box;
  }

  return NULL;

} /* leftist */

t_boxlik *rightist(t_boxlik *box)
{
  if( box == NULL )
  {
    return NULL;
  }

  ret = leftist(box->trd);
  if( ret != NULL)
  {
    return ret;
  }

  ret = leftist(box->snd);
  if( ret != NULL)
  {
    return ret;
  }
  
  ret = leftist(box->fst);
  if( ret != NULL)
  {
    return ret;
  }

  if( box->typ == BX_TXT )
  {
    return box;
  }

  return NULL;

} /* leftist */

#else

t_boxlik *leftist(t_boxlik *box)
{
  t_boxlik *next= NULL;

       if( box->fst ) next = box->fst;
  else if( box->snd ) next = box->snd;
  else if( box->trd ) next = box->trd;
   
  if( next ) return leftist(next);
  else return box;

} /* leftist */

t_boxlik *rightist(t_boxlik *box)
{
  t_boxlik *next= NULL;

       if( box->trd ) next = box->trd;
  else if( box->snd ) next = box->snd;
  else if( box->fst ) next = box->fst;
   
  if( next ) return rightist(next);
  else return box;

} /* rightist */
#endif

void
print_boxlik_pp_mezi(t_boxlik *le, t_boxlik *ri)
{
  le = rightist(le);
  ri = leftist(ri);

  if ( le == NULL || ri == NULL )
  {
    return;
  }

  for(le = le->next; le != NULL && le != ri; le = le -> next )
  {
    print_boxlik_pp_one(le);
  }
} /* print_boxlik_pp_mezi */

void
print_boxlik_pp(t_boxlik *box)
{
  print_boxlik_pp_one(box);

  if( box->typ == BX_DECLARATION)
  {
    return;
  }

  if( box->fst)
  {
    print_boxlik_pp(box->fst);
  }

  if( box-> fst && box->snd )
  {
    print_boxlik_pp_mezi(box->fst, box->snd);
  }

  if( box->snd)
  {
    print_boxlik_pp(box->snd);
  }
  else if ( box->fst && box->trd)
  {
    print_boxlik_pp_mezi(box->fst, box->trd);
  }

  if( box-> snd && box->trd )
  {
    print_boxlik_pp_mezi(box->snd, box->trd);
  }

  if( box->trd)
  {
    print_boxlik_pp(box->trd);
  }
} /* print_boxlik_pp */


void 
print_boxlik_pp_all()
{
   print_boxlik_pp_one(fst_boxlik);
   print_boxlik_pp_mezi(fst_boxlik, boxlik_root);
   print_boxlik_pp(boxlik_root);
   print_boxlik_pp_mezi(boxlik_root, lst_boxlik);
   print_boxlik_pp_one(lst_boxlik);
} /*  print_boxlik_pp_all */


/***************************
 * 
 * olejuj
 * 
 * hlavni ologatova cast, pusti se do narasovaneho syntaktickeho stromu
 * ve kterem jsou expresny maji vyhodnocene typy
 *
 * kdyz ve stromu narazi na nastaveny BX_LOGAT_BIT
 * zmeni binarni operaci na logatovou fci 
 * tj pred hodi (napr) "lg_add(", za hodi ")" a misto operatoru da ","
 * vse ostatni necha.
 *
 * kdyz narazi na `cast' jine nez BX_NIC ohranici (narp) "lg_ltof(" a ")"
 * 
 */

void olejuj_boxlik(t_boxlik *box);
void olejuj_doleva(t_boxlik *ss, char *txt);
void olejuj_doprava(t_boxlik *ss, char *txt);
void print_boxlik_linear(int);
void olejuj_carku(t_boxlik *box);

char * logat_name_table[] =
{
#if 0
 "lg_inc_post(",
 "lg_dec_post(",
 "lg_inc_pref(",
 "lg_dec_pref(",

 "lg_arg_list???(",
 "lg_expr???(",
 "lg_unary???(",
 "lg_sizeof???(",
 "lg_sizeof???(",
 "lg_cast???(",

  "lg_mul(",
  "lg_div(",
  "lg_mod(",
  "lg_add(",
  "lg_sub(",
 "bx_inc_pref",
 "bx_un_exlaim",
 "bx_un_tilda",
 "bx_inc_post",
 "bx_dec_pref",
 "bx_dec_post",
 "bx_un_plus",
 "bx_un_minus",
 "bx_cast",
 "bx_asterix",
 "bx_obelix",
 "bx_sizeof_expr",
 "bx_sizeof_type",
#endif
 "lg_not(",
 "lg_neg(",
 "lg_inc_pref(",
 "lg_inc_post(",
 "lg_dec_pref(",
 "lg_dec_post(",
 "lg_un_plus(",
 "lg_un_minus(",
 "lg_cast(",
 "lg_asterix(",
 "lg_obelix(",
 "lg_sizeof(",
 "lg_sizeof(",

 "lg_fce(",
 "lg_array(",
 "lg_ptr_op(",
 "lg_dot(",

 "lg_mul(",
 "lg_div(",
 "lg_mod(",
 "lg_add(",
 "lg_sub(",
 "lg_lshift(",
 "lg_rshift(",
 "lg_lt(",
 "lg_le(",
 "lg_gt(",
 "lg_ge(",
 "lg_eq(",
 "lg_neq(",
 "lg_and(",
 "lg_xor(",
 "lg_or(",
 "lg_land(",
 "lg_lor(",

 "lg_assig(",
 "lg_assig_eq(",
 "lg_assig_add(",
 "lg_assig_sub(",
 "lg_assig_mul(",
 "lg_assig_div(",
 "lg_assig_mod(",
 "lg_assig_left(",
 "lg_assig_right(",
 "lg_assig_and(",
 "lg_assig_xor(",
 "lg_assig_or(",
 "lg_comma(",
 "lg_cond(",

};


/*
 * na zaklade `typ'u vrati nazev vce ktera se ma pridat 
 */
char*
olejuj_logat_name(int typ)
{
  //assert(typ & BX_LOGAT_BIT );

  typ &= ~BX_LOGAT_BIT;

  assert( typ >= BX_UN_EXLAIM);
  assert( typ <= BX_COND);

  return logat_name_table[typ - BX_UN_EXLAIM];
} /* olejuj_all */


/*
 * podobne jako `olejuj_logat_name()' akorat ze vraci jmeno pretypovavaci fce
 */

char*
olejuj_cast_name(int cast)
{
  assert(cast > BX_NIC);
  assert(cast <= BX_DTOL);

  return bx_nazvy_cast[cast];
} /* olejuj_cast_name */

/*
 * frontend k olejuj_boxlik
 */
void
olejuj_all()
{
  olejuj_boxlik(boxlik_root);
  print_boxlik_linear(0);
} /* olejuj_all */


/*
 * rekurzivne projde syntakticky strom a naolejuje vsechny expresiony
 */
void
olejuj_boxlik(t_boxlik *box)
{
  int typ;

  if( box == NULL )
  {
    return;
  }

  /* rekurizvne */
  olejuj_boxlik(box->fst);
  olejuj_boxlik(box->snd);
  olejuj_boxlik(box->trd);

  /* rozeber si co se dela s timto nodikem */
  typ = box ->typ;

  if( typ & BX_LOGAT_BIT )
  {
    
    typ &= ~BX_LOGAT_BIT;

    if( typ >= BX_MUL &&  typ <= BX_COMMA )
    {
      olejuj_doleva(box, olejuj_logat_name(box->typ) );
      olejuj_carku(box);
      olejuj_doprava(box, ")/*LOGAT BINOP*/");
    }
    else if( typ >= BX_UN_EXLAIM && typ <= BX_SIZEOF_TYPE )
    {
      /* olejuj unarni operator */
    }
  }
  
#if 1
  if( box -> cast !=  BX_NIC )
  {
    olejuj_doleva(box, olejuj_cast_name(box->cast) );
    olejuj_doprava(box, ")/*LOGAT CAST*/");
  }
#endif
 

}


/*
 * pridej text `txt' do leva podstromu `ss' ale tak aby pak byl v podstromu
 * `ss' 
 * to je ponekdu zahackovane: vytvori se nove dva listy do
 * leveho se da `txt' do praveho se prekopiruej puvodni nejlevjejsi list
 * a jeho txt se zmeni na ""
 * cele se to prislusne propji pomoci spojaku `next'
 */
void
olejuj_doleva(t_boxlik *ss, char *txt)
{
  t_boxlik *ri, *nwfst, *nwsnd;

  assert(ss);

  ri = leftist(ss);

  assert(ri->typ == BX_TXT || ri->typ == BX_IDENTIFIER);
  assert(!ri->fst&&!ri->snd&&!ri->trd);
  assert(ri->next);

  nwfst = boxlik_new_txt(txt);
  nwsnd = BX0(BX_TXT);

  ri->fst = nwfst;
  ri->snd = nwsnd;

  nwsnd->next = ri->next;
  nwfst->next = nwsnd;
  ri->next = nwfst;

  nwsnd->txt = ri->txt;
  ri->txt = "";

} /* olejuj_doleva */

/*
 * pridej do stromu `ss' text `txt' 
 * podobne opicarny jako v olejuj_doprava 
 */

void
olejuj_doprava(t_boxlik *ss, char *txt)
{
  t_boxlik *ri, *nwfst, *nwsnd;

  assert(ss);

  ri = rightist(ss);

  assert(ri->typ == BX_TXT || ri->typ == BX_IDENTIFIER);
  assert(!ri->fst&&!ri->snd&&!ri->trd);
  assert(ri->next);

  nwfst = BX0(BX_TXT);
  nwsnd = boxlik_new_txt(txt);

  ri->fst = nwfst;
  ri->snd = nwsnd;

  nwsnd->next = ri->next;
  nwfst->next = nwsnd;
  ri->next = nwfst;
  
  
  nwfst->txt =  ri->txt;
  ri->txt = "";

} /* olejuj_doprava */

/*
 * misto "+" je treba dat ","
 */

void
olejuj_carku(t_boxlik *box)
{
  assert(box);
  assert(box->snd);
  assert(box->snd->typ == BX_TXT);
  assert(box->snd->txt);

  /* FIXME: dirty hack */
  box->snd->txt = ",";
} /* olejuj_carku */

/*****************************
 *
 *  linear
 *
 *  a nyni neni nic slozitejsiho nez tento spojak vytisknout. jupiii
 */

/*
 * vytisken spojak boxliku (spojenych `next') resp jejich `txt'
 * kdyz bde `drawbox' != 0 nakresli se boxliky [asi]-[ ]-[tahovehle]-[
 * ]-[aby]-[" bylo videt, jak je to naboxlikovne"]
 *
 */

void
print_boxlik_linear(int drawbox)
{
  t_boxlik *i;

  for(i= fst_boxlik; i; i= i->next)
  {
    if(i->txt)
    {
      bx_printf(drawbox?"-[%s]":"%s", i->txt);
    }
    else
    {
      bx_err("print_boxlik_linear: NULL == i->next )");
    }
  }
} /* print_boxlik_linear */

/*
 * usage
 */ 

void usage()
{
  printf(
    "logat [-sltvh][-p prefile][-o outputfile] infile\n"
    "    -o outputfile\n"
    "    -p prefile, (silent declarations)\n"
    "    -l print linear boxliks\n"
    "    -h this help\n\n"
    "    -s suppres printig #include \"logop.h\"\n"
    "   debug options:\n"
    "    -v print var table\n"
    "    -t tree parse tree\n"
    "    -l print linear boxliks\n"
    );
} /* usage */

/*
 * yyin -- vstupni soubor pouziva flex
 */
extern FILE* yyin;

/*
 * main
 */
int
main(int argc, char *argv[])
{
  int tree = 0;
  int linear = 0;
  int zn;
  char *out_file_name = NULL;
  char *pre_file_name = NULL;
  FILE *f;
  FILE *prefile = NULL; 
  int prepend_include = 1;

  t_boxlik *i;


  bx_file_out = stdout;
  bx_file_err = stderr;

  while ( (zn = getopt(argc, argv, "p:o:htvsl")) != -1 )
  {
    switch(zn)
    {
    case 'v': print_var_table = 1; break;
    case 't': tree = 1; break;
    case 'l': linear = 1; break;
    case 'o': out_file_name = optarg; break;
    case 'p': pre_file_name = optarg; break;
    case 's': prepend_include = 0; break;
    case 'h':
    default : usage(); return  0; 
    
    }
  }

  argc -= optind;
  argv += optind;

  type_tbl = idtable_new();
  type_tbl_loc = idtable_new();

  /* predsoubor */
  if( pre_file_name )
  {
    yyin = fopen(pre_file_name, "rt");
    if( yyin == NULL)
    {
      bx_err("cannt fopen(%s)\n", pre_file_name);
      return -1;
    }

    /* ehm ponekud dirty hack (strom z presouboru zustane vyset jen tak */
    yyparse();

    fclose(yyin);
    yyin = stdin;

    fst_boxlik = NULL;
    lst_boxlik = NULL;
    boxlik_root = NULL;
    
  }

  /* presmerovani vstupniho souboru */
  if( out_file_name )
  {
    bx_file_out = fopen(out_file_name, "wt");
    if( f == NULL)
    {
      bx_err("cannt fopen(%s)\n", out_file_name);
      return -1;
    }
    
  }

  /* vstupni soubor */
  if(argc >= 1 )
  {
    yyin = fopen(argv[0], "r");
    if( yyin == NULL)
    {
      bx_err("cannt fopen(%s)\n", argv[0]);
      return -1;
    }
   
  }

  yyparse();
  
  if(print_var_table)
  {
    bx_printf("---== type_tbl ==---\n\n");
    idtable_print(type_tbl);
  }

  if(linear)
  {
    bx_printf("---== print_boxlik_linear ==---\n\n");
    print_boxlik_linear(1);
  }

  assert(boxlik_root != NULL);


  if( tree )
  {
    bx_printf("---== print_boxlik_tree ==---\n\n");
    print_boxlik_tree(0, boxlik_root, 0, '#');
  }

  if( prepend_include )
  {
    bx_printf("#include \"logop.h\"\n\n");
  }

  olejuj_all();



  return 0;

} /* main */



